//Note: this is a logic based agent, but I did not use the given knowledge base
//It seemed like it was optional from the wording so I used my own if that is okay

//Testing 
//import java.io.FileNotFoundException;
//import java.io.PrintWriter;

//Necessary 
import java.util.Set;
import java.util.Stack;
import king.Agent;

//Might need 
//import java.util.Random;
//import java.awt.Point;
//import java.lang.Math;

public class RhogarAgent extends Agent {
	
	
	//Internal members
	
	//Action lastAction : lastAction is automatically updated with the last action returned by your GetAction().
	//					  On your first turn, lastAction is set to Action.NULL
	//                    
	//Point facing : facing is a vector representing the direction you're facing.
	//               You start off facing down. The book states you can't ask the world your direction,
	//               so you must update it yourself when you choose to turn.
	//               TurnedLeft() and TurnedRight() are helper methods for updating your direction.
	//               up: <0,-1>, down: <0,1>, left: <-1,0>, right: <1,0>
	//
	//Point position : position is a vector representing the position of your agent. Its up to you to change it
	//                 when your agent successfully moves forward (no bump percept). MovedForward() will do this for you.
		
	public RhogarAgent() {
		//Give your agent a unique name
		super("Rhogar");
		
		for(int i = 0; i < 10; i++)
		{
			for(int j = 0; j < 10; j++)
			{
				dungeonMap[i][j] = State.UNEXPLORED;
			}
		}
	}
	
	public enum State{
		EXPLORED, FRONTIER, UNSAFE, UNEXPLORED, MAYBE_WUMPUS, MAYBE_PIT;
	}
	
	State[][] dungeonMap = new State[11][11];
	
	Stack<Action> safePath = new Stack<Action>();
	public Set<Percept> currPercept;
	//position current position
	//facing direction
	int rightTurns = 0;
	boolean hasGold,nowhereSafe,turnedAround,wumpusDead, hasArrow, turnedLeft = false;
	
	//Each run through a Wumpus Room is a series of update loops.
	//Each loop, the world first calls your agent's PercieveResult method, and gives it a set of percepts
	//based on the current tile you're on. The percepts are from the Percept Enum, and are STENCH,BREEZE,BUMP,GLITTER,SCREAM.
	//It is up to you what you do with them.
	//The world then calls GetAction(). Based on the series of percepts you've been given, you must return the most
	//appropriate Action. The Action Enum contains TURN_LEFT, TURN_RIGHT, FORWARD, SHOOT, PICK_UP, CLIMB.
	
	//The world is asking what your agent wants to do.
	@Override
	public Action GetAction() {
		
		if(hasGold == true || nowhereSafe == true)
		{
			return escape();
		}
		else if(currPercept.contains(Percept.GLITTER))
		{
			hasGold = true;
			return Action.PICK_UP;
		}
		else if(lastAction == Action.NULL )
		{
			return firstMove();
		}
		else if(lastAction == Action.SHOOT) {
			if(currPercept.contains(Percept.SCREAM)) {
				
				markFrontier();
				wumpusDead = true;
				
				if(turnedLeft)
				return rightMove();
				else	
				return downMove();
			}
			else {
				
				nowhereSafe = true;
				return escape();
			}
		}
		else
		{
			if(turnedLeft)
				return rightMove();
				else	
				return downMove();
		}
	}
	
	//The percepts set tells your agent about the tile he's currently on.
	@Override
	public void PerceiveResult(Set<Percept> percepts) {
			
		currPercept = percepts;
	}
	
	//With nowhere safe or gold in hand, escape to surface through safe path
	public Action escape() {
		
		if(safePath.isEmpty() || (position.x==1 && position.y==1))
		{
			return Action.CLIMB;
		}
		else if(safePath.peek() == Action.FORWARD &&(!turnedAround))
		{
			if((rightTurns == 0 || rightTurns == 1))
			{
				rightTurns += 1;
				TurnedRight();
				return Action.TURN_RIGHT;
			}
			else if(rightTurns == 2)
			{
				
				rightTurns = 0;
				turnedAround = true;
				return escape();
			}
		}
		else if(safePath.peek() == Action.TURN_LEFT)
		{
			safePath.pop();
			TurnedRight();
			return Action.TURN_RIGHT;
		}
		else if(safePath.peek() == Action.TURN_RIGHT)
		{
			safePath.pop();
			TurnedLeft();
			return Action.TURN_LEFT;
		}
			
		safePath.pop();
		MovedForward();
		return Action.FORWARD;
		
	}
	
	//Makes varying first moves depending on state of first tile
	public Action firstMove() {
		
		dungeonMap[position.x - 1][position.y - 1] = State.EXPLORED;
		markFrontier();
		
		if(currPercept.contains(Percept.BREEZE))
		{
			markPit();
			return Action.CLIMB;
		}
		else if(currPercept.contains(Percept.STENCH))
		{
			markWumpus();
			return Action.SHOOT;
		}
		
		MovedForward();
		dungeonMap[position.x - 1][position.y - 1] = State.EXPLORED;
		markFrontier();
		safePath.push(Action.FORWARD);
		return Action.FORWARD;	
	}	
	
	//Makes various moves after first move
	public Action downMove()
	{	
		if(currPercept.contains(Percept.BUMP) && dungeonMap[position.x][position.y-1] == State.FRONTIER)
		{	
			dungeonMap[position.x - 1][position.y - 1] = State.UNSAFE;
			dungeonMap[position.x][position.y - 1] = State.UNSAFE;
			position.y = position.y - 1;
			TurnedLeft();
			turnedLeft = true;
			safePath.push(Action.TURN_LEFT);
			return Action.TURN_LEFT;
		}
		else if(currPercept.contains(Percept.BUMP))
		{	
			dungeonMap[position.x - 1][position.y - 1] = State.UNSAFE;
			dungeonMap[position.x][position.y - 1] = State.UNSAFE;
			position.y = position.y - 1;
			nowhereSafe = true;
			return escape();
		}
		else if(currPercept.contains(Percept.BREEZE))
		{
			markPit();
			nowhereSafe = true;
			return escape();
		}
		else if(currPercept.contains(Percept.STENCH))
		{
			if(!wumpusDead)
			{
				markWumpus();
				if(hasArrow = true) {
					hasArrow = false;
					return Action.SHOOT;
				}
				
				nowhereSafe = true;
				return escape();
			}
		}
		
		MovedForward();
		dungeonMap[position.x - 1][position.y - 1] = State.EXPLORED;
		markFrontier();
		safePath.push(Action.FORWARD);
		return Action.FORWARD;
	}
	
	public Action rightMove()
	{	
		if(currPercept.contains(Percept.BUMP))
		{	
			dungeonMap[position.x - 1][position.y - 1] = State.UNSAFE;
			dungeonMap[position.x][position.y - 1] = State.UNSAFE;
			position.y = position.y - 1;
			nowhereSafe = true;
			return escape();
		}
		else if(currPercept.contains(Percept.BREEZE))
		{
			markPit();
			nowhereSafe = true;
			return escape();
		}
		else if(currPercept.contains(Percept.STENCH))
		{
			if(!wumpusDead)
			{
				markWumpus();
				if(hasArrow = true) {
					hasArrow = false;
					return Action.SHOOT;
				}
				
				nowhereSafe = true;
				return escape();
			}
		}
		
		position.x = position.x + 1;
		dungeonMap[position.x - 1][position.y - 1] = State.EXPLORED;
		safePath.push(Action.FORWARD);
		return Action.FORWARD;
	}
	
	public void markPit()
	{
		dungeonMap[position.x][position.y-1] = State.MAYBE_PIT;
		dungeonMap[position.x-1][position.y] = State.MAYBE_PIT;
	}
	public void markWumpus()
	{
		dungeonMap[position.x][position.y-1] = State.MAYBE_WUMPUS;
		dungeonMap[position.x-1][position.y] = State.MAYBE_WUMPUS;
	}
	public void markFrontier()
	{
		dungeonMap[position.x][position.y - 1] = State.FRONTIER;
	}
}
