import java.awt.Point;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;

import king.Agent;
import king.Tile;


public class ExtendedAgent extends Agent {
	private int turn;
	private Map<Point, Tile> tiles;
	private Tile currentTile;
	public Action previousAction = Action.NULL; 
	private Set<Percept> CurrentPercepts;
	//Internal members
	
	//Action lastAction : lastAction is automatically updated with the last action returned by your GetAction().
	//					  On your first turn, lastAction is set to Action.NULL
	//                    
	//Point facing : facing is a vector representing the direction you're facing.
	//               You start off facing down. The book states you can't ask the world your direction,
	//               so you must update it yourself when you choose to turn.
	//               TurnedLeft() and TurnedRight() are helper methods for updating your direction.
	//               up: <0,-1>, down: <0,1>, left: <-1,0>, right: <1,0>
	//
	//Point position : position is a vector representing the position of your agent. Its up to you to change it
	//                 when your agent successfully moves forward (no bump percept). MovedForward() will do this for you.
	

	public ExtendedAgent() {
		//Give your agent a unique name
		super("Your Agent Name Here");
		this.tiles = new HashMap<Point, Tile>();
		this.turn = 0;
	}
	
	
	//Each run through a Wumpus Room is a series of update loops.
	//Each loop, the world first calls your agent's PercieveResult method, and gives it a set of percepts
	//based on the current tile you're on. The percepts are from the Percept Enum, and are STENCH,BREEZE,BUMP,GLITTER,SCREAM.
	//It is up to you what you do with them.
	//The world then calls GetAction(). Based on the series of percepts you've been given, you must return the most
	//appropriate Action. The Action Enum contains TURN_LEFT, TURN_RIGHT, FORWARD, SHOOT, PICK_UP, CLIMB.
	
	
	@Override
	public Action GetAction() {
		//The world is asking what your agent wants to do.
		this.turn += 1; //Always keep track of what turn it is.
		Action currentAction = Action.TURN_LEFT;
		//Smooth tiles map based on percepts
		if(this.CurrentPercepts.contains(Percept.BUMP))
			this.bump();
		if(this.CurrentPercepts.contains(Percept.BREEZE))
			this.smoothPit();
		else if(this.CurrentPercepts.contains(Percept.STENCH))
			this.smoothWumpus();
		else if(!this.tiles.get(this.position).explored)
			this.smoothSafe();
		
		//Decide on next action
		if(this.CurrentPercepts.contains(Percept.GLITTER)){
			return Action.PICK_UP;
		}
		else if(!this.tiles.containsKey(this.facing)) {
//				currentAction = Action.FORWARD;
//				this.MovedForward();
		}
		else if(!this.tiles.get(this.facing).explored){
			if(!this.tiles.get(this.facing).pit && !this.tiles.get(this.facing).wumpus){
//					currentAction = Action.FORWARD;
//					this.MovedForward();
			}
			else{
				currentAction = Action.TURN_LEFT;
				this.TurnedLeft();
			}
		}
		else{
			currentAction = Action.TURN_LEFT;
			this.TurnedLeft();
		}
		
		//Upkeep actions and return statement
		this.turn += 1;
		this.previousAction = this.lastAction;
		this.lastAction = currentAction;
		return Action.CLIMB;
	}
	public void bump(){
		Tile bumpTile = new Tile();
		bumpTile.pit = true;
		this.tiles.put(this.position, bumpTile);
	}
	public boolean comparePoints(Point A, Point B){
		return A.getX() == B.getX() && A.getY() == B.getY();
	}
	@Override
	public void PerceiveResult(Set<Percept> percepts) {
		//The percepts set tells your agent about the tile he's currently on.
		this.CurrentPercepts = percepts; //Let agent know what is going on in the current tile.
		this.currentTile = new Tile();
		if(this.CurrentPercepts.contains(Percept.BREEZE)) this.currentTile.breeze = true;
		if(this.CurrentPercepts.contains(Percept.GLITTER)) this.currentTile.gold = true;
		if(this.CurrentPercepts.contains(Percept.STENCH)) this.currentTile.stench = true;
		this.currentTile.explored = true;
		this.tiles.put(this.position, this.currentTile);	
	}	
	
	//This handles if a breeze is seen
	public void smoothPit(){
		Point right = new Point(this.position.x + 1, this.position.y);
		Point left = new Point(this.position.x - 1, this.position.y);
		Point up = new Point(this.position.x, this.position.y + 1);
		Point down = new Point(this.position.x, this.position.y - 1);
		Tile pitTile = new Tile();
		pitTile.pit = true;
		//Smooth right
		if(this.tiles.containsKey(right)){
			if(this.tiles.get(right).explored == false)
				this.tiles.get(right).pit = true;
		}
		else {
			this.tiles.put(right, new Tile(pitTile));
		}
		//Smooth left
		if(this.tiles.containsKey(left)){
			if(this.tiles.get(left).explored == false)
				this.tiles.get(left).pit = true;
		}
		else {
			this.tiles.put(left, new Tile(pitTile));
		}
		//Smooth up
		if(this.tiles.containsKey(up)){
			if(this.tiles.get(up).explored == false)
				this.tiles.get(up).pit = true;
		}
		else {
			this.tiles.put(up, new Tile(pitTile));
		}
		//Smooth down
		if(this.tiles.containsKey(down)){
			if(this.tiles.get(down).explored == false)
				this.tiles.get(down).pit = true;
		}
		else {
			this.tiles.put(down, new Tile(pitTile));
		}
	}
	//This handles if a stench is seen
	public void smoothWumpus() {
		Point right = new Point(this.position.x + 1, this.position.y);
		Point left = new Point(this.position.x - 1, this.position.y);
		Point up = new Point(this.position.x, this.position.y + 1);
		Point down = new Point(this.position.x, this.position.y - 1);
		Tile wumpusTile = new Tile();
		wumpusTile.wumpus = true;
		//Smooth right
		if(this.tiles.containsKey(right)){
			if(!this.tiles.get(right).explored)
				this.tiles.get(right).wumpus = true;
		}
		else {
			this.tiles.put(right, new Tile(wumpusTile));
		}
		//Smooth left
		if(this.tiles.containsKey(left)){
			if(!this.tiles.get(left).explored)
				this.tiles.get(left).wumpus = true;
		}
		else {
			this.tiles.put(left, new Tile(wumpusTile));
		}
		//Smooth up
		if(this.tiles.containsKey(up)){
			if(!this.tiles.get(up).explored )
				this.tiles.get(up).wumpus = true;
		}
		else {
			this.tiles.put(up, new Tile(wumpusTile));
		}
		//Smooth down
		if(this.tiles.containsKey(down)){
			if(!this.tiles.get(down).explored)
				this.tiles.get(down).wumpus = true;
		}
		else {
			this.tiles.put(down, new Tile(wumpusTile));
		}
	}
	
	//This handles if there are no hazzards
	public void smoothSafe(){
		Point right = new Point(this.position.x + 1, this.position.y);
		Point left = new Point(this.position.x - 1, this.position.y);
		Point up = new Point(this.position.x, this.position.y + 1);
		Point down = new Point(this.position.x, this.position.y - 1);
		Tile safeTile = new Tile();
		safeTile.pit = false;
		safeTile.wumpus = false;
		//Smooth right
		if(this.tiles.containsKey(right)){
			if(this.tiles.get(right).explored == false){
				this.tiles.get(right).pit = false;
				this.tiles.get(right).wumpus = false;
			}
		}
		else {
			this.tiles.put(right, new Tile(safeTile));
		}
		//Smooth left
		if(this.tiles.containsKey(left)){
			if(this.tiles.get(left).explored == false)
			{
				this.tiles.get(right).pit = false;
				this.tiles.get(right).wumpus = false;
			}
		}
		else {
			this.tiles.put(left, new Tile(safeTile));
		}
		//Smooth up
		if(this.tiles.containsKey(up)){
			if(this.tiles.get(up).explored == false)
			{
				this.tiles.get(right).pit = false;
				this.tiles.get(right).wumpus = false;
			}
		}
		else {
			this.tiles.put(up, new Tile(safeTile));
		}
		//Smooth down
		if(this.tiles.containsKey(down)){
			if(this.tiles.get(down).explored == false)
			{
				this.tiles.get(right).pit = false;
				this.tiles.get(right).wumpus = false;
			}
		}
		else {
			this.tiles.put(down, new Tile(safeTile));
		}
	}
}