package king.logical;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;


public class ClauseCNF {
	
	private Map<String, Boolean> map;
	
	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + ((map == null) ? 0 : map.hashCode());
		for(Map.Entry<String,Boolean> literal : map.entrySet()){
			result = prime * result + literal.getKey().hashCode()*(literal.getValue()?prime-1:prime-15);
		}
		return result-result+31;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		ClauseCNF other = (ClauseCNF) obj;
		if (map == null) {
			if (other.map != null)
				return false;
		} else {
			for(Map.Entry<String,Boolean> literal : map.entrySet()){
				if(!other.map.containsKey(literal.getKey())||literal.getValue()!=other.map.get(literal.getKey())){
					return false;
				}
			}
			
		}
		return true;
	}

	public ClauseCNF(){
		map = new TreeMap<String,Boolean>();
	}
	
	public ClauseCNF(String s, boolean b){
		map = new TreeMap<String,Boolean>();
		map.put(s, b);
	}

	public ClauseCNF merge(ClauseCNF other){
		//Set<String> contra = new TreeSet<String>();
		String contra = "";
		
		for(Map.Entry<String, Boolean> literal : map.entrySet()){
			if(other.map.containsKey(literal.getKey())&&other.map.get(literal.getKey())!=literal.getValue()){
				if(!contra.equals(""))
					return null;
				contra = literal.getKey();
			}
		}
		if(contra.equals(""))
			return null;
		
		ClauseCNF temp = new ClauseCNF();
		temp.map.putAll(this.map);
		temp.map.putAll(other.map);
		temp.map.remove(contra);
		
		return temp;
	}
	
	public ClauseCNF unsafeMerge(ClauseCNF other){  //faster, but no checking for unmergability
		for(Map.Entry<String, Boolean> literal : this.map.entrySet()){
			if(other.map.containsKey(literal.getKey())&&other.map.get(literal.getKey())!=literal.getValue()){
				ClauseCNF temp = new ClauseCNF();
				temp.map.putAll(this.map);
				temp.map.putAll(other.map);
				temp.map.remove(literal.getKey());
				return temp;
			}
		}
		return null;
	}
	
	public boolean canMerge(ClauseCNF other){
		int count = 0;
		for(Map.Entry<String, Boolean> literal : this.map.entrySet()){
			if(other.map.containsKey(literal.getKey())&&other.map.get(literal.getKey())!=literal.getValue())
				++count;
		}
		return count == 1;
	}
	
	public static boolean Mergable(ClauseCNF a, ClauseCNF b){
		return a.canMerge(b);
	}
	
	public void add(String key, Boolean value){
		map.put(key, value);
	}
	
	public int size(){
		return map.size();
	}
	
	public String toString(){
		return map.toString();
	}
	
	protected boolean absorb(ClauseCNF other){
		for(Map.Entry<String, Boolean> literal : other.map.entrySet()){
			if(this.map.containsKey(literal.getKey())&&this.map.get(literal.getKey())!=literal.getValue())
				return false;
			map.put(literal.getKey(), literal.getValue());
		}
		return true;
	}
	
	protected List<ClauseCNF> deMorgans(){
		List<ClauseCNF> dmorg = new ArrayList<ClauseCNF>();
		for(Map.Entry<String, Boolean> literal : this.map.entrySet())
			dmorg.add(new ClauseCNF(literal.getKey(),!literal.getValue()));
		return dmorg;
	}
	
}
