package king.logical;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Stack;


public class KnowledgeBase {
	private KnowledgeBaseCNF cnfKB = new KnowledgeBaseCNF();
	
	
	public KnowledgeBase(){
		
	}
	
	public void tell(LogicalObject sentence){
		for(ClauseCNF clause : sentence.convertToCNF())
		if(!cnfKB.contains(clause))
			cnfKB.add(clause);
	}
	
	public boolean ask(String symbol, Boolean value){
		KnowledgeBaseCNF.verbose = false;
		return cnfKB.Resolution(symbol, value);
	}
	
	public boolean askVerbose(String symbol, Boolean value){
		KnowledgeBaseCNF.verbose = true;
		return cnfKB.Resolution(symbol, value);
	}
	
	public void setVerbose(boolean value){
		KnowledgeBaseCNF.verbose = value;
	}
	
	/*
	public void tell(String str){
		String[] tokens = str.split(" ");
		Stack<LogicalObject> stack = new Stack<LogicalObject>();
		for(int i = tokens.length-1; i >= 0; --i){
			if(tokens[i].equals("AND")){
				stack.push(Logic.conjunction(new LogicalObject[]{stack.pop(),stack.pop()}));
			}else if(tokens[i].equals("OR")){
				stack.push(Logic.conjunction(new LogicalObject[]{stack.pop(),stack.pop()}));
			}
		}
	}
	*/
	
	public void tell(String str){
		//Map<Integer,Integer> parenMap = new HashMap<Integer,Integer>();
		//if(!mapParenthesis(str,parenMap))
		//	return;
		//String[] tokens = str.split(" ");
		
		
		Map<Integer,Integer> parenMap = new HashMap<Integer,Integer>();
		str = str.replaceAll("\\(", " ( ");
		str = str.replaceAll("\\)" , " ) ");
		String [] tokens = str.split("\\s+");
		if(!mapParenthesis(tokens,parenMap))
			return;
		try{
			if(tokens[0].equals("(")){
				if(!tokens[tokens.length-1].equals(")"))
					throw new ParseException("");
				LogicalObject lo = parse(tokens,1,tokens.length-2,parenMap);
				tell(lo);
			}else{
				LogicalObject lo = parse(tokens,0,tokens.length-1,parenMap);
				tell(lo);
			}
		}catch(ParseException e){
			System.out.println("Not a valid sentence: " + str);
		}
	}
	
	private boolean mapParenthesis(String[] tokens,Map<Integer,Integer> map){
		Stack<Integer> stack = new Stack<Integer>();
		for(int i = 0; i < tokens.length; ++i){
			if(tokens[i].equals("(")){
				stack.push(i);
			}else if(tokens[i].equals(")")){
				if(stack.empty())
					return false;
				map.put(stack.pop(), i);
			}
		}
		return stack.empty();
	}
	
	private class ParseException extends Exception{
		private static final long serialVersionUID = 5610339737841259068L;

		public ParseException(String message){
			super(message);
		}
	}
	
	private LogicalObject parse(String[] tokens, int start, int end, Map<Integer,Integer> parens) throws ParseException{
		if(end < start)
			throw new ParseException("");
		LogicalObject.Type type = interpretType(tokens[start]);
		if(type == null)
			return toLiteral(tokens[start]);
		List<LogicalObject> list = new ArrayList<LogicalObject>();
		for(int i = start + 1; i <= end;){
			if(tokens[i].equals("(")){
				list.add(parse(tokens,i+1,parens.get(i)-1,parens));
				i = parens.get(i) + 1;
			}else{
				list.add(toLiteral(tokens[i]));
				++i;
			}
		}
		LogicalObject[] larr = new LogicalObject[list.size()];
		if(list.isEmpty())
			throw new ParseException("Bad Logic String");
		if(type == LogicalObject.Type.Disjunction)
			return Logic.disjunction(list.toArray(larr));
		if(type == LogicalObject.Type.Conjunction)
			return Logic.conjunction(list.toArray(larr));
		if(list.size() != 2)
			throw new ParseException("Bad Logic String");
		if(type == LogicalObject.Type.Implication)
			return Logic.implication(list.toArray(larr));
		//must be biconditional
		return Logic.biconditional(list.toArray(larr));
	}
	
	private LogicalObject toLiteral(String str){
		if(str.charAt(0)=='~')
			return Logic.literal(str.substring(1),false);
		return Logic.literal(str);
	}
	
	private LogicalObject.Type interpretType(String token){
		if(token.equals("AND") || token.equals("and") || token.equals("^"))
			return LogicalObject.Type.Conjunction;
		if(token.equals("OR") || token.equals("or") || token.equals("V"))
			return LogicalObject.Type.Disjunction;
		if(token.equals("->"))
			return LogicalObject.Type.Implication;
		if(token.equals("<->"))
			return LogicalObject.Type.Biconditional;
		return null;
	}
}
