package king;

import java.awt.Point;
import java.util.Random;
import java.util.Set;
import java.util.TreeSet;

public class World {
	
	public enum State{RUNNING,DEATH_BY_WUMPUS,DEATH_BY_PIT,WON,ESCAPED,OUT_OF_TURNS};
	public boolean terminal = false;
	public State state = State.RUNNING;
	
	public Tile[][] tiles;
	public Point agentPos = new Point(0,0);
	public Point agentDir = new Point(0,1);
	public int height, width;
	private Random rand = new Random();
	public Agent agent = new DummyAgent();
	public boolean hasArrow = true;
	public boolean hasGold = false;
	private int score = 0;
	private int turns = 0;
	public static final int MAX_TURNS = 1000;
	
	public int getScore(){
		return score;
	}
	
	public Point getAgentPos(){
		return (Point) agentPos.clone();
	}
	
	public World(){
		GenWorld(5,5,0.1f);
	}
	
	private World(int w, int h, Tile[][] copytiles){
		width = w;
		height = h;
		tiles = new Tile[w][h];
		for(int i = 0; i < w; ++i){
			for(int j = 0; j < h; ++j){
				tiles[i][j] = new Tile(copytiles[i][j]);
			}
		}
	}
	
	public World Copy(Agent a){
		World w = new World(width,height,tiles);
		w.agent = a;
		w.agent.PerceiveResult(w.getPerceptsFromTile(w.agentPos));
		return w;
	}
	
	public World(int w, int h){
		GenWorld(w,h,0.2f);
	}
	
	public World(int w, int h, Agent a){
		agent = a;
		GenWorld(w,h,0.2f);
		//agent.PerceiveResult(getPerceptsFromTile(agentPos));
	}
	
	public World(int w, int h, float pitR){
		GenWorld(w,h,pitR);
	}
	
	private void GenWorld(int w, int h, float pitR){
		width = w;
		height = h;
		tiles = new Tile[width][height];
		for(int i = 0; i < w; ++i){
			for(int j = 0; j < h; ++j){
				tiles[i][j] = new Tile();
			}
		}
		for(int i = 0; i < w; ++i){
			for(int j = 0; j < h; ++j){
				if(rand.nextFloat()<pitR)
					placePit(i,j);
			}
		}
		while(!placeWumpus(rand.nextInt(w),rand.nextInt(h))){}
		while(!placeGold(rand.nextInt(w),rand.nextInt(h))){}
		tiles[0][0].explored = true;
		agent.PerceiveResult(getPerceptsFromTile(agentPos));
	}
	
	private boolean inBounds(int x, int y,boolean intro){
		return (intro||x!=0||y!=0) && x >= 0 && x < width && y >= 0 && y < height;
	}
	
	private boolean placePit(int x, int y){
		if(!inBounds(x,y,false))
			return false;
		
		tiles[x][y].pit = true;
		placeBreeze(x-1,y);
		placeBreeze(x+1,y);
		placeBreeze(x,y-1);
		placeBreeze(x,y+1);
		return true;
	}
	
	private boolean placeBreeze(int x, int y){
		if(!inBounds(x,y,true))
			return false;
		
		tiles[x][y].breeze = true;
		return true;
	}
	
	private boolean placeWumpus(int x, int y){
		if(!inBounds(x,y,false))
			return false;
		
		tiles[x][y].wumpus = true;
		tiles[x][y].stench = true;
		placeStench(x-1,y);
		placeStench(x+1,y);
		placeStench(x,y-1);
		placeStench(x,y+1);
		return true;
	}
	
	private boolean placeStench(int x, int y){
		if(!inBounds(x,y,true))
			return false;
		
		tiles[x][y].stench = true;
		return true;
	}
	
	private boolean placeGold(int x, int y){
		if(!inBounds(x,y,false))
			return false;
		
		tiles[x][y].gold = true;
		return true;
	}
	
	public void tick(){
		synchronized(this){
		if(state != State.RUNNING)
			return;
		if(++turns>MAX_TURNS){
			state = State.OUT_OF_TURNS;
			return;
		}
		score -= 1;
		Agent.Action action = agent.GetAction();
		agent.lastAction = action;
		Set<Agent.Percept> percepts = new TreeSet<Agent.Percept>();
		if(action == Agent.Action.FORWARD){
			Point dest = (Point) agentPos.clone();
			dest.translate(agentDir.x,agentDir.y);
			if(!inBounds(dest.x,dest.y,true))
				percepts.add(Agent.Percept.BUMP);
			else{
				agentPos = dest;
				tiles[agentPos.x][agentPos.y].explored = true;
				if(tiles[agentPos.x][agentPos.y].wumpus){
					score = -1000;
					state = State.DEATH_BY_WUMPUS;
					return;
				}
				else if(tiles[agentPos.x][agentPos.y].pit){
					score = -1000;
					state = State.DEATH_BY_PIT;
					return;
				}
			}
		}else if(action == Agent.Action.TURN_RIGHT){
			agentDir = new Point(-agentDir.y,agentDir.x);
		}else if(action == Agent.Action.TURN_LEFT){
			agentDir = new Point(agentDir.y,-agentDir.x);
		}else if(action == Agent.Action.PICK_UP){
			if(tiles[agentPos.x][agentPos.y].gold){
				tiles[agentPos.x][agentPos.y].gold = false;
				hasGold = true;
			}
		}else if(action == Agent.Action.SHOOT){
			if(hasArrow){
				score -= 10;
				hasArrow = false;
				for(Point apos = (Point)agentPos.clone();inBounds(apos.x, apos.y, true);apos.translate(agentDir.x,agentDir.y)){
					if(tiles[apos.x][apos.y].wumpus){
						percepts.add(Agent.Percept.SCREAM);
						tiles[apos.x][apos.y].wumpus = false;
						tiles[apos.x][apos.y].deadwumpus = true;
						break;
					}
				}
			}
		}else if(action == Agent.Action.CLIMB){
			if(agentPos.x == 0 && agentPos.y == 0){
				state = hasGold?State.WON:State.ESCAPED;
				if(state == State.WON)
					score += 1000;
				return;
			}
		}
		percepts.addAll(getPerceptsFromTile(agentPos));
		agent.PerceiveResult(percepts);
		}
	}
	
	private Set<Agent.Percept> getPerceptsFromTile(Point p){
		Set<Agent.Percept> set = new TreeSet<Agent.Percept>();
		if(!inBounds(p.x,p.y,true))
			return set;
		Tile t = tiles[p.x][p.y];
		if(t.breeze)
			set.add(Agent.Percept.BREEZE);
		if(t.gold)
			set.add(Agent.Percept.GLITTER);
		if(t.stench)
			set.add(Agent.Percept.STENCH);
		return set;
	}
}