// Author: Sean King

package king;

import java.awt.Color;
import java.awt.Graphics;
import java.awt.Image;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ComponentEvent;
import java.awt.event.ComponentListener;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.awt.image.BufferedImage;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.imageio.ImageIO;
import javax.swing.BoxLayout;
import javax.swing.JButton;
import javax.swing.JDialog;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;
import javax.swing.JPanel;
import javax.swing.JSpinner;
import javax.swing.SpinnerNumberModel;

import king.logical.KnowledgeBase;
import king.logical.Logic;
import king.logical.LogicalObject;

public class WumpusGUI extends JFrame implements KeyListener {
	
	public static final int TILE_SIZE = 70;
	public static final int PADDING = 10;
	
	public List<AgentFactory> factoryList = new ArrayList<AgentFactory>();
	
	public static AgentFactory factory = null;
	public JMenuBar menubar;
	public JMenu filemenu;
	
	Thread wumpusLoop = null;
	boolean terminateLoop = false;
	

	Thread tourneyLoop = null;
	boolean terminateTourneyLoop = false;
	
	private JPanel drawPanel;
	private BufferedImage drawbuffer = new BufferedImage(500,500,BufferedImage.TYPE_INT_ARGB);
	public World world;
	
	public static class Images{
		public static Image guy,wumpus,gold,breeze,pit,stench,tile,exploredtile,deadwumpus,endingwumpus,endingpit,endinggold,endingnothing,endingturnlimit,coverart;
	}

	/**
	 * 
	 */
	private static final long serialVersionUID = -5549863196167703522L;

	public static void main(String [] args){
		WumpusGUI app = new WumpusGUI();
	}
	
	public WumpusGUI(){
		super("Wumpus World");
		createHeader();
		loadResources();
		
		
		this.setSize(300,300);
		this.setDefaultCloseOperation(EXIT_ON_CLOSE);
		
		
		//world = new World(4,4);
		drawbuffer = new BufferedImage(300,300,BufferedImage.TYPE_INT_ARGB);
		drawbuffer.getGraphics().drawImage(Images.coverart, 0,0, null);
		
		
		//drawbuffer.getGraphics().
		
		drawPanel = new JPanel();
		
		drawPanel.setSize(300,300);
		add(drawPanel);
		this.addComponentListener(new ComponentListener(){
			@Override
			public void componentHidden(ComponentEvent arg0) {}
			@Override
			public void componentMoved(ComponentEvent e) {}
			@Override
			public void componentResized(ComponentEvent e) {}
			@Override
			public void componentShown(ComponentEvent e) {
				renderworld();
			}
		});
		drawPanel.addComponentListener(new ComponentListener(){
			@Override
			public void componentHidden(ComponentEvent arg0) {}
			@Override
			public void componentMoved(ComponentEvent e) {}
			@Override
			public void componentResized(ComponentEvent e) {
				renderworld();
			}
			@Override
			public void componentShown(ComponentEvent e) {}
		});
		
		this.addKeyListener(this);
		this.setFocusable(true);
		
		this.setVisible(true);
		
		wumpusLoop = new Thread(){
			public void run(){
				
				while(!terminateLoop){
					try {
						Thread.sleep(100);
					} catch (InterruptedException e) {
						e.printStackTrace();
					}
					if(filemenu.isSelected())
						continue;
					renderworld();
				}
				
			}
		};
		wumpusLoop.start();
	}
	
	private void createHeader(){
		menubar = new JMenuBar();
		filemenu = new JMenu("File");
		filemenu.setMnemonic('f');
		JMenuItem newitem = new JMenuItem("New World");
		newitem.setMnemonic('n');
		newitem.addActionListener(new ActionListener(){

			@Override
			public void actionPerformed(ActionEvent e) {
				// TODO Auto-generated method stub
				newDialog();
			}
			
		});
		filemenu.add(newitem);
		
		
		JMenuItem tournyItem = new JMenuItem("Tournament");
		
		tournyItem.setMnemonic('t');
		tournyItem.addActionListener(new ActionListener(){

			@Override
			public void actionPerformed(ActionEvent arg0) {
				// TODO Auto-generated method stub
				tourneyDialog();
			}
		});
		filemenu.add(tournyItem);
		
		
		menubar.add(filemenu);
		this.setJMenuBar(menubar);
	}
	
	//Much of the code for this method was recycled from ConnectK
	private void newDialog(){
		final JDialog jd = new JDialog();
		JLabel widthLabel = new JLabel("Width:");
		JLabel heightLabel = new JLabel("Height:");
		final JSpinner width = new JSpinner(new SpinnerNumberModel(4,2,10,1));
		final JSpinner height = new JSpinner(new SpinnerNumberModel(4,2,10,1));
		JPanel panel = new JPanel();
		panel.setLayout(new BoxLayout(panel, BoxLayout.X_AXIS));
		panel.add(widthLabel);
		panel.add(width);
		panel.add(heightLabel);
		panel.add(height);
		JButton nbutton = new JButton("New World");
		nbutton.addActionListener(new ActionListener(){

			@Override
			public void actionPerformed(ActionEvent arg0) {
				terminateTourneyLoop = true;
				if(tourneyLoop != null)
					try {
						tourneyLoop.join();
						tourneyLoop = null;
					} catch (InterruptedException e1) {
						// TODO Auto-generated catch block
						e1.printStackTrace();
					}
				terminateTourneyLoop = false;
				CreateWorld((Integer)width.getValue(),(Integer)height.getValue());
				jd.dispose();
			}
			
		});
		panel.add(nbutton);
		
		final JFrame pf = this;
		final JFileChooser fc = new javax.swing.JFileChooser(new java.io.File("."));
		fc.addActionListener(new ActionListener(){
			@Override
			public void actionPerformed(ActionEvent arg0) {
				java.io.File file = ((JFileChooser) arg0.getSource()).getSelectedFile();
				factory = new AgentFactory(file);
			}
		});
		JButton addAI = new JButton("Load AI Class");
		addAI.addActionListener(new ActionListener(){
			@Override
			public void actionPerformed(ActionEvent arg0) {
				fc.showOpenDialog(pf);
			}
		});
		panel.add(addAI);
		
		
		//panel.setSize(400,400);
		jd.add(panel);
		jd.setSize(400, 70);
		jd.setVisible(true);
	}
	
	private void tourneyDialog(){
		
		final JFrame pf = this;
		final JDialog jd = new JDialog();
		JLabel widthLabel = new JLabel("Width:");
		JLabel heightLabel = new JLabel("Height:");
		JLabel roundsLabel = new JLabel("Rounds: ");
		final JSpinner width = new JSpinner(new SpinnerNumberModel(4,2,10,1));
		final JSpinner height = new JSpinner(new SpinnerNumberModel(4,2,10,1));
		final JSpinner rounds = new JSpinner(new SpinnerNumberModel(1,1,10000,1));
		JPanel panel = new JPanel();
		panel.setLayout(new BoxLayout(panel, BoxLayout.X_AXIS));
		panel.add(widthLabel);
		panel.add(width);
		panel.add(heightLabel);
		panel.add(height);
		panel.add(roundsLabel);
		panel.add(rounds);
		JButton nbutton = new JButton("Select Tournament Folder");
		nbutton.addActionListener(new ActionListener(){

			@Override
			public void actionPerformed(ActionEvent arg0) {
				
				final JFileChooser fc = new javax.swing.JFileChooser(new java.io.File("."));
				fc.addActionListener(new ActionListener(){
					@Override
					public void actionPerformed(ActionEvent arg0) {
						factoryList.clear();
						//java.io.File file = ((JFileChooser) arg0.getSource()).getSelectedFile();
						java.io.File folder = ((JFileChooser) arg0.getSource()).getCurrentDirectory();
						for(java.io.File file : folder.listFiles()){
							if(!file.isDirectory() && AgentFactory.isAgentFile(file)){
								AgentFactory af = new AgentFactory(file);
								if(af.isValid())
									factoryList.add(new AgentFactory(file));
							}
						}
						if(!factoryList.isEmpty()){
							runTourney((Integer)width.getValue(), (Integer)height.getValue(),(Integer)rounds.getValue());
						}
					}
				});
				fc.showOpenDialog(pf);
				jd.dispose();
			}
			
		});
		panel.add(nbutton);
		
		jd.add(panel);
		jd.setSize(500, 70);
		jd.setVisible(true);
	}
	
	public class Result{
		public String name;
		public int score;
		Result(String n, int s){
			name = n;
			score = s;
		}
	}
	public Map<String, Result> resultMap = new HashMap<String, Result>();
	
	
	public int tw = 4, th = 4, tr = 1;
	private void runTourney(int w, int h, int runs){
		terminateTourneyLoop = true;
		if(tourneyLoop != null)
			try {
				tourneyLoop.join();
			} catch (InterruptedException e1) {
				// TODO Auto-generated catch block
				e1.printStackTrace();
			}
		terminateTourneyLoop = false;
		
		
		tw = w;
		th = h;
		tr = runs;
		tourneyLoop = new Thread(){
			Map<String, Result> resultmap = new HashMap<String, Result>();
			List<Result> resultlist = new ArrayList<Result>();
			public void run(){
				java.io.File tfile = new File("tournament_results.txt");
				java.io.PrintStream ps = null;
				try {
					ps = new java.io.PrintStream(tfile);
				} catch (FileNotFoundException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
					return;
				}
				for(int i = 0; i < tr; ++i){
					World model = new World(tw,th);
					for(AgentFactory af : factoryList){
						factory = af;
						CreateWorld(model);
						while(world.state==World.State.RUNNING)
							try {
								if(terminateTourneyLoop){
									return;
								}
								Thread.sleep(100);
							} catch (InterruptedException e) {
								// TODO Auto-generated catch block
								e.printStackTrace();
							}
						
						if(resultmap.containsKey(world.agent.name)){
							resultmap.get(world.agent.name).score += world.getScore();
						}else{
							resultmap.put(world.agent.name, new Result(world.agent.name,world.getScore()));
							resultlist.add(resultmap.get(world.agent.name));
						}
						//resultlist.add(new Result(world.agent.name,world.getScore()));
						ps.println(world.agent.name+" "+world.getScore());
					}
				}
				Collections.sort(resultlist,new Comparator<Result>(){
					@Override
					public int compare(Result a, Result b) {
						return b.score-a.score;
					}
				});
				int lastscore = 99999999; //impossible score
				int place = 1;
				for(int i = 0; i < resultlist.size(); ++i){
					if(resultlist.get(i).score != lastscore){
						lastscore = resultlist.get(i).score;
						place = i+1;
					}
					ps.println(place+". " + resultlist.get(i).name+" ~ " + resultlist.get(i).score);
				}
				ps.close();
				final JDialog jd = new JDialog();
				jd.add(new JLabel("Tournament Finished"));
				jd.setSize(200, 100);
				jd.setVisible(true);
			}
		};
		tourneyLoop.start();
	}
	
	public void loadResources(){
		try {
			Images.guy = ImageIO.read(WumpusGUI.class.getResource("/guy.png"));
			Images.wumpus = ImageIO.read(WumpusGUI.class.getResource("/livewumpus.png"));
			Images.gold = ImageIO.read(WumpusGUI.class.getResource("/gold.png"));
			Images.breeze = ImageIO.read(WumpusGUI.class.getResource("/breeze.png"));
			Images.pit = ImageIO.read(WumpusGUI.class.getResource("/pit.png"));
			Images.stench = ImageIO.read(WumpusGUI.class.getResource("/stench2.png"));
			Images.tile = ImageIO.read(WumpusGUI.class.getResource("/biggertile.png"));
			Images.deadwumpus = ImageIO.read(WumpusGUI.class.getResource("/deadwumpus.png"));
			Images.exploredtile = ImageIO.read(WumpusGUI.class.getResource("/exploredtile.png"));
			Images.endingwumpus = ImageIO.read(WumpusGUI.class.getResource("/KilledByWumpus.png"));
			Images.endingnothing = ImageIO.read(WumpusGUI.class.getResource("/LeftWithNothing.png"));
			Images.endinggold = ImageIO.read(WumpusGUI.class.getResource("/GotTheGold.png"));
			Images.endingpit = ImageIO.read(WumpusGUI.class.getResource("/KilledByPit.png"));
			Images.endingturnlimit = ImageIO.read(WumpusGUI.class.getResource("/TooManyMoves.png"));
			Images.coverart = ImageIO.read(WumpusGUI.class.getResource("/wumpuscover.png"));
		} catch (IOException e) {
			e.printStackTrace();
		}
	}
	
	public void flipBuffer(){
		//drawPanel.getGraphics().drawImage(drawbuffer, 0,0,null);
		drawPanel.getGraphics().drawImage(drawbuffer, 0, 0, drawPanel.getWidth(), drawPanel.getHeight(), 0, 0, TILE_SIZE*world.width, TILE_SIZE*world.height, null);
		drawbuffer.getGraphics().setColor(Color.WHITE);
		drawbuffer.getGraphics().fillRect(0, 0, 500, 500);
	}
	
	public void clearBuffer(){
		drawbuffer.getGraphics().setColor(Color.WHITE);
		drawbuffer.getGraphics().fillRect(0, 0, 500, 500);
	}
	
	public void renderworld(){
		if(world == null){
			drawPanel.getGraphics().drawImage(Images.coverart,0,0,drawPanel.getWidth(),drawPanel.getHeight(),0,0,300,300,null);
			return;
		}
		synchronized(world){
		draw(world);
		flipBuffer();
		}
	}
	
	public void draw(World w){
		Graphics g = drawbuffer.getGraphics();
		for(int i = 0; i < w.width; ++i){
			for(int j = 0; j < w.height; ++j){
				Tile t = w.tiles[i][j];
				g.drawImage(t.explored?Images.exploredtile:Images.tile,i*TILE_SIZE,j*TILE_SIZE,null);
				if(t.pit)
					g.drawImage(Images.pit, i*TILE_SIZE+PADDING,j*TILE_SIZE+PADDING,null);
				if(t.wumpus)
					g.drawImage(Images.wumpus,i*TILE_SIZE+PADDING,j*TILE_SIZE+PADDING,null);
				if(t.deadwumpus)
					g.drawImage(Images.deadwumpus,i*TILE_SIZE+PADDING,j*TILE_SIZE+PADDING,null);
				if(t.gold)
					g.drawImage(Images.gold, i*TILE_SIZE+PADDING,j*TILE_SIZE+PADDING,null);
				if(t.breeze && !t.wumpus && !t.pit)
					g.drawImage(Images.breeze, i*TILE_SIZE+PADDING,j*TILE_SIZE+PADDING,null);
				if(t.stench && !t.wumpus && !t.pit && !t.deadwumpus)
					g.drawImage(Images.stench, i*TILE_SIZE+PADDING,j*TILE_SIZE+PADDING,null);
			}
		}
		g.drawImage(Images.guy, world.getAgentPos().x*TILE_SIZE+PADDING, world.getAgentPos().y*TILE_SIZE+PADDING, null);
		if(w.state == World.State.DEATH_BY_WUMPUS){
			g.drawImage(Images.endingwumpus, drawbuffer.getWidth()/2-100, drawbuffer.getHeight()/2-100, null);
		}else if(w.state == World.State.DEATH_BY_PIT){
			g.drawImage(Images.endingpit, drawbuffer.getWidth()/2-100, drawbuffer.getHeight()/2-100, null);
		}else if(w.state == World.State.ESCAPED){
			g.drawImage(Images.endingnothing, drawbuffer.getWidth()/2-100, drawbuffer.getHeight()/2-100, null);
		}else if(w.state == World.State.WON){
			g.drawImage(Images.endinggold, drawbuffer.getWidth()/2-100, drawbuffer.getHeight()/2-100, null);
		}else if(w.state == World.State.OUT_OF_TURNS){
			g.drawImage(Images.endingturnlimit, drawbuffer.getWidth()/2-100, drawbuffer.getHeight()/2-100, null);
		}
	}
	
	public void CreateWorld(int w,int h){
		//this.setSize(w, h);
		terminateLoop = true;
		if(wumpusLoop != null)
			try {
				wumpusLoop.join();
			} catch (InterruptedException e) {
				e.printStackTrace();
			}
		terminateLoop = false;
		
		world = (factory==null)?(new World(w,h)):(new World(w,h,factory.generateAgent()));
		drawbuffer = new BufferedImage(TILE_SIZE*world.width,TILE_SIZE*world.height,BufferedImage.TYPE_INT_ARGB);
		renderworld();
		this.setSize(w*100,h*100);
		
		wumpusLoop = new Thread(){
			public void run(){
				
				while(!terminateLoop){
					
					
					try {
						Thread.sleep(100);
					} catch (InterruptedException e) {
						e.printStackTrace();
					}
					if(filemenu.isSelected())
						continue;
					if(world.state == World.State.RUNNING)
						world.tick();
					renderworld();
				}
				
			}
		};
		wumpusLoop.start();
	}
	
	private void CreateWorld(World model){
		//this.setSize(w, h);
		terminateLoop = true;
		if(wumpusLoop != null)
			try {
				wumpusLoop.join();
			} catch (InterruptedException e) {
				e.printStackTrace();
			}
		terminateLoop = false;
		
		world = model.Copy(factory.generateAgent());
		drawbuffer = new BufferedImage(TILE_SIZE*world.width,TILE_SIZE*world.height,BufferedImage.TYPE_INT_ARGB);
		renderworld();
		this.setSize(model.width*100,model.height*100);
		
		wumpusLoop = new Thread(){
			public void run(){
				
				while(!terminateLoop){
					
					
					try {
						Thread.sleep(100);
					} catch (InterruptedException e) {
						e.printStackTrace();
					}
					if(filemenu.isSelected())
						continue;
					if(world.state == World.State.RUNNING)
						world.tick();
					renderworld();
				}
				
			}
		};
		wumpusLoop.start();
	}

	@Override
	public void keyPressed(KeyEvent eve) {
		// TODO Auto-generated method stub
		if(eve.getKeyCode()==KeyEvent.VK_R){
			terminateTourneyLoop = true;
			if(tourneyLoop != null)
				try {
					tourneyLoop.join();
				} catch (InterruptedException e1) {
					// TODO Auto-generated catch block
					e1.printStackTrace();
				}
			terminateTourneyLoop = false;
			CreateWorld(4,4);
		}else if(eve.getKeyCode()==KeyEvent.VK_F){
			world.tick();
			renderworld();
		}
	}

	@Override
	public void keyReleased(KeyEvent arg0) {
		// TODO Auto-generated method stub
		
	}

	@Override
	public void keyTyped(KeyEvent arg0) {
		// TODO Auto-generated method stub
		
	}
	
	
	
	
}
