import java.util.Set;
import king.logical.KnowledgeBase;
import king.Agent;
import java.util.ArrayList;
import java.awt.Point;
import java.lang.Math;

public class MyAgent extends Agent {
	
	
	//Internal members
	
	//Action lastAction : lastAction is automatically updated with the last action returned by your GetAction().
	//					  On your first turn, lastAction is set to Action.NULL
	//                    
	//Point facing : facing is a vector representing the direction you're facing.
	//               You start off facing down. The book states you can't ask the world your direction,
	//               so you must update it yourself when you choose to turn.
	//               TurnedLeft() and TurnedRight() are helper methods for updating your direction.
	//               up: <0,-1>, down: <0,1>, left: <-1,0>, right: <1,0>
	//
	//Point position : position is a vector representing the position of your agent. Its up to you to change it
	//                 when your agent successfully moves forward (no bump percept). MovedForward() will do this for you.
	
	public Set<Percept> p;
	private ArrayList<Action> toDo;
	private ArrayList<Point> empty;
	private ArrayList<ArrayList<Double>> dangerRate;
	private boolean hasArrow;
	private boolean isDead;
	private boolean hasGold;
	private int height = 4;
	private int width = 4;
	

	public MyAgent() {
		//Give your agent a unique name
		super("Team AD");
		toDo = new ArrayList<Action>();
		empty = new ArrayList<Point>();
		dangerRate = new ArrayList<ArrayList<Double>>();
		for(int i = 0; i < width; ++i)
			dangerRate.add(new ArrayList<Double>());
//		for(int i = 1; i <= width; ++i)
//			for(int j = 1; j <= height; ++j)
//				if(i == 1 && j == 1)
//					dangerRate.get(i).add(0.0);
//				else
//					dangerRate.get(i).add(0.2);
		hasArrow = true;
		isDead = false;
		hasGold = false;
	}
	
	
	//Each run through a Wumpus Room is a series of update loops.
	//Each loop, the world first calls your agent's PercieveResult method, and gives it a set of percepts
	//based on the current tile you're on. The percepts are from the Percept Enum, and are STENCH,BREEZE,BUMP,GLITTER,SCREAM.
	//It is up to you what you do with them.
	//The world then calls GetAction(). Based on the series of percepts you've been given, you must return the most
	//appropriate Action. The Action Enum contains TURN_LEFT, TURN_RIGHT, FORWARD, SHOOT, PICK_UP, CLIMB.
	
	
	@Override
	public Action GetAction() {
		//The world is asking what your agent wants to do.
		if(hasGold && position.x == 1 && position.y == 1)
			return Action.CLIMB;
		if(toDo.size() > 0){
			Action toReturn = toDo.get(0);
			toDo.remove(0);
			return toReturn;
		}
		if(p.size() == 0)
			empty.add(position);
		for(Percept i : p){
			switch(i){
				case STENCH:{
					stench();
					break;
				}
				case BREEZE:{
					breeze();
					break;
				}
				case BUMP:{
					bump();
					break;
				}
				case GLITTER:{
					hasGold = true;
					return Action.PICK_UP;
					//break;
				}
				case SCREAM:{
					scream();
					break;
				}
			}
		}
		double minRate = 0;
		if(position.y <= height - 1)
			return goTo(new Point(position.x, position.y+1));
		else if(position.y >= 2)
			return goTo(new Point(position.x, position.y-1));
		return Action.FORWARD;
	}

	private void stench(){

	}
	
	private void breeze(){
//		if(position.x <= width - 1)
//			dangerRate.get(position.x).set((position.y)-1,0.25);
//		if(position.x - 1 >= 1)
//			dangerRate.get(position.x-2).set((position.y)-1, 0.25);
//		if(position.y <= height - 1)
//			dangerRate.get(position.x-1).set(position.y, 0.25);
//		if(position.y - 1 >= 1)
//			dangerRate.get(position.x-1).set((position.y)-2, 0.25);
	}

	private void bump(){
		if (facing.x == 0 && facing.y == 1)
			height = position.y;
		else if(facing.x == 1 && facing.y == 0)
			width = position.x;
	}

	private void glitter(){

	}

	private void scream(){
		isDead = true;
	}

	private Action goTo(Point destination){
		int X = destination.x - position.x;
		int Y = destination.y - position.y;
		if(facing.x == X && facing.y == Y)
			toDo.add(Action.FORWARD);
		else if(Math.abs(facing.x - X) == 2 || Math.abs(facing.y - Y) == 2){
			toDo.add(Action.TURN_LEFT);
			toDo.add(Action.TURN_LEFT);
			toDo.add(Action.FORWARD);
		}
		else
			toDo.add(Action.FORWARD);
		Action toReturn = toDo.get(0);
		toDo.remove(0);
		return toReturn;
	}

	@Override
	public void PerceiveResult(Set<Percept> percepts) {
		//The percepts set tells your agent about the tile he's currently on.
		p.clear();
		for(Percept i : percepts)
			p.add(i);
	}

	
}
