import java.util.Set;
import java.util.Stack;
import java.util.HashSet;
import king.Agent;
import java.util.Map;
import java.util.Random;
import java.util.HashMap;
import java.util.List;
import java.util.Arrays;
import java.util.Collections;
import java.awt.Point;

class Danger{
	public static int NONE = 5, UNKNOWN = 0, LOW = 25, MEDIUM = 50, HIGH = 75, VERYHIGH = 100, IMPOSSIBLE = 500; //Assume unexplored tiles are a bit dangerous
	public static int value = UNKNOWN;
}

enum Direction{LEFT, RIGHT, UP, DOWN, NULL}

class TileInfo{
	public boolean stench;
	public boolean breeze;
	public int personalrating;
	public Point point;
	public TileInfo(Point p, boolean stench, boolean breeze){
		point = p;
		this.stench = stench;
		this.breeze = breeze;
		personalrating = Danger.NONE;
	};
	int RelativeRating(){ //Depends only on what is in the tile
		int rating = 0;
		if (stench)
			rating += 20;
		if (breeze)
			rating += 25;
		return rating;
	}

	int PersonalRating(){
		return personalrating;
	}
}

class ProbabilityKB{

	public Map<Integer, TileInfo> tileKB = new HashMap<Integer, TileInfo>();
	
	public ProbabilityKB(){}
	
	public void AddData(Point p, TileInfo ti){
		if (!tileKB.containsKey(makeHashCode(p))){
			tileKB.put(makeHashCode(p), ti);
		}else{
			tileKB.get(makeHashCode(p)).personalrating += 5; //Rating increases on each additional visit
		}
	}
	
	 public static int makeHashCode(Point p) {
		 int x = p.x;
		 int y = p.y;
		 int hash = 17;
        hash = ((hash + x) << 5) - (hash + x);
        hash = ((hash + y) << 5) - (hash + y);
        return hash;
	}
	
	public int DangerRating(Point pt, int width, int height)
	{
		int rating = Danger.UNKNOWN;
		if (tileKB.containsKey(makeHashCode(pt))){
			rating += tileKB.get(makeHashCode(pt)).PersonalRating();
		}else{
			Set<Point> adjacentPoints = new HashSet<Point>();
			adjacentPoints.add(new Point(pt.x - 1, pt.y));
			adjacentPoints.add(new Point(pt.x + 1, pt.y));
			adjacentPoints.add(new Point(pt.x, pt.y - 1));
			adjacentPoints.add(new Point(pt.x, pt.y + 1));
			for (Point point : adjacentPoints){
				if (point.x == 0 || point.y ==0 || point.x > width || point.y > height){continue;} //Invalid points
				if (tileKB.containsKey(makeHashCode(point)))
				{
					rating += tileKB.get(makeHashCode(point)).RelativeRating();
				}else{
					rating += Danger.UNKNOWN;
				}
			}
		}
		return rating;
	}
	
}
public class TsalakoProbabilityAgent extends Agent {
	private Set<Percept> percSet;
	private ProbabilityKB KB;
	private Direction recommendedDirection = Direction.NULL;
	private Point lastRecPosition = new Point(-1, -1);
	private int width = 11; //Assume max + 1
	private int height = 11;
	private Stack<Point> previousSteps = new Stack<Point>();
	private boolean gotGold = false;
	//Internal members
	
	//Action lastAction : lastAction is automatically updated with the last action returned by your GetAction().
	//					  On your first turn, lastAction is set to Action.NULL
	//                    java
	//Point facing : facing is a vector representing the direction you're facing.
	//               You start off facing down. The book states you can't ask the world your direction,
	//               so you must update it yourself when you choose to turn.
	//               TurnedLeft() and TurnedRight() are helper methods for updating your direction.
	//               up: <0,-1>, down: <0,1>, left: <-1,0>, right: <1,0>
	//
	//Point position : position is a vector representing the position of your agent. Its up to you to change it
	//                 when your agent successfully moves forward (no bump percept). MovedForward() will do this for you.
	

	public TsalakoProbabilityAgent() {
		//Give your agent a unique name
		super("Secret Agent Tsalako");
		KB = new ProbabilityKB();

	}
	
	
	//Each run through a Wumpus Room is a series of update loops.
	//Each loop, the world first calls your agent's PercieveResult method, and gives it a set of percepts
	//based on the current tile you're on. The percepts are from the Percept Enum, and are STENCH,BREEZE,BUMP,GLITTER,SCREAM.
	//It is up to you what you do with them.
	//The world then calls GetAction(). Based on the series of percepts you've been given, you must return the most
	//appropriate Action. The Action Enum contains TURN_LEFT, TURN_RIGHT, FORWARD, SHOOT, PICK_UP, CLIMB.
	
	
	@Override
	public Action GetAction() {
		//The world is asking what your agent wants to do.
		Action action;
		
		action = gotGold ? RecommendActionWithGold() : RecommendAction();
		
		return action;
	}

	@Override
	public void PerceiveResult(Set<Percept> percepts) {
		percSet = percepts;
		
		if (this.lastAction == Action.FORWARD && !percepts.contains(Percept.BUMP)){
			this.MovedForward();
		}else if(percepts.contains(Percept.BUMP)){
			Direction face = GetFaceDirection();
			if (face == Direction.RIGHT){
				width = this.position.x;
			}else if(face == Direction.DOWN){
				height = this.position.y ;
			}
		}
		
		if (this.position == new Point(1,1) || (this.lastAction == Action.FORWARD && !percepts.contains(Percept.BUMP))){
			TileInfo tileinfo = new TileInfo(this.position, percepts.contains(Percept.STENCH), percepts.contains(Percept.BREEZE));
			KB.AddData(new Point(this.position), tileinfo);
			if (!gotGold && this.lastAction == Action.FORWARD )
				previousSteps.push((Point) this.position.clone());
			else if (gotGold && !previousSteps.isEmpty())
				previousSteps.pop();
		}

			
	}
	
	public Direction RecommendDirection(){
		Point pt = new Point(this.position);
		Random rand = new Random();
		int[] adjacentDR = new int[]{Danger.IMPOSSIBLE, Danger.IMPOSSIBLE, Danger.IMPOSSIBLE, Danger.IMPOSSIBLE}; //Assume the worst
		Direction[] adjacentDirection = new Direction[]{Direction.NULL, Direction.NULL, Direction.NULL, Direction.NULL};
		Set<Integer> bestIndices = new HashSet<Integer>();
		if (pt.x > 1){
			adjacentDR[0] = KB.DangerRating(new Point(pt.x - 1, pt.y), width, height);
			adjacentDirection[0] = Direction.LEFT;
		}
		if (pt.x < width){
			adjacentDR[1] = KB.DangerRating(new Point(pt.x + 1, pt.y), width, height);
			adjacentDirection[1] = Direction.RIGHT;
		}
		if (pt.y > 1){
			adjacentDR[2] = KB.DangerRating(new Point(pt.x, pt.y - 1), width, height);
			adjacentDirection[2] = Direction.UP;
		}
		if (pt.y < height){
			adjacentDR[3] = KB.DangerRating(new Point(pt.x, pt.y + 1), width, height);
			adjacentDirection[3] = Direction.DOWN;
		}

		String dir[] = new String[]{"LEFT", "RIGHT", "UP", "DOWN"};

		int bestIndex = 0;
		for (int i = 0; i < 4; i++){
			if (adjacentDR[i] < adjacentDR[bestIndex]){
				bestIndex = i;
				bestIndices.clear(); //Reset the set if a better Danger Rating is found.
			}else if (adjacentDR[i] == adjacentDR[bestIndex]){
				bestIndices.add(i);
			}
		}
		bestIndices.add(bestIndex);
		int r = rand.nextInt(bestIndices.size());
		int i = 0;
		int selectedIndex = bestIndex;
		for (int index : bestIndices){
			if (i == r){
				selectedIndex = index;
			}
			i += 1;
		}
		return adjacentDirection[selectedIndex];
			
	}
	
	public Action RecommendActionWithGold()
	{
		
		if (this.position.x == 1 && this.position.y == 1){
			return Action.CLIMB;
		}else if (!previousSteps.isEmpty()){
			Point p = previousSteps.peek();
			if (this.position.x == p.x && this.position.y == p.y)
				previousSteps.pop(); //Remove the current position we are in.
		}

		
		Point goal = previousSteps.isEmpty() ? new Point(1,1) : previousSteps.peek();
		Direction facing = GetFaceDirection();
		Direction recommendedDirection = RecommendDirectionToPoint(goal);
		if (recommendedDirection == facing){
			return Action.FORWARD;
		}else{
			return FaceDirection(recommendedDirection);
		}
	}
	
	public Action RecommendAction(){
		if (percSet.contains(Percept.GLITTER)) //Gold state, pick up gold
		{
			gotGold = true;
			return Action.PICK_UP;
		}
		
		if (recommendedDirection == Direction.NULL){
			recommendedDirection = RecommendDirection();
		}
		Direction facing = GetFaceDirection();
		if (recommendedDirection == facing){
			recommendedDirection = Direction.NULL;
			return Action.FORWARD;
		}else {
			return FaceDirection(recommendedDirection);
		}
	}
	
	public Direction GetFaceDirection(){
		Point facing = new Point(this.facing);
		if (facing.equals(new Point(-1, 0))){
			return Direction.LEFT;
		}
		else if (facing.equals(new Point(1, 0))) {
			return Direction.RIGHT;
		}
		else if (facing.equals(new Point(0, -1))) {
			return Direction.UP;
		}
		else {return Direction.DOWN;}
	}
	
	public Direction RecommendDirectionToPoint(Point p){ //Based on the goal Point, recommend a direction we should face.
		if (this.position.x < p.x){
			return Direction.RIGHT;
		}else if(this.position.x > p.x){
			return Direction.LEFT;
		}else if(this.position.y < p.y){
			return Direction.DOWN;
		}else if(this.position.y > p.y){
			return Direction.UP;
		}else{
			return Direction.NULL;
		}
	}
	
	public Action FaceDirection(Direction d){ //Face the given direction
		//Will always turn right if we have equal options
		switch(GetFaceDirection()){
		case LEFT:
			if (d == Direction.RIGHT){
				this.facing = new Point(0, -1);
				return Action.TURN_RIGHT;
			}else if (d == Direction.UP){
				this.facing = new Point(0, -1);
				return Action.TURN_RIGHT;
			}else if (d == Direction.DOWN){
				this.facing = new Point(0, 1);
				return Action.TURN_LEFT;
			}
			break;
		case RIGHT:
			if (d == Direction.LEFT){
				this.facing = new Point(0, 1);
				return Action.TURN_RIGHT;
			}else if (d == Direction.UP){
				this.facing = new Point(0, -1);
				return Action.TURN_LEFT;
			}else if (d == Direction.DOWN){
				this.facing = new Point(0, 1);
				return Action.TURN_RIGHT;
			}
			break;
		case UP:
			if (d == Direction.LEFT){
				this.facing = new Point(-1, 0);
				return Action.TURN_LEFT;
			}else if (d == Direction.RIGHT){
				this.facing = new Point(1, 0);
				return Action.TURN_RIGHT;
			}else if (d == Direction.DOWN){
				this.facing = new Point(1, 0);
				return Action.TURN_RIGHT;
			}
			break;
		case DOWN:
			if (d == Direction.LEFT){
				this.facing = new Point(-1, 0);
				return Action.TURN_RIGHT;
			}else if (d == Direction.RIGHT){
				this.facing = new Point(1, 0);
				return Action.TURN_LEFT;
			}else if (d == Direction.UP){
				this.facing = new Point(-1, 0);
				return Action.TURN_RIGHT;
			}
			break;
		}
		return Action.NULL;
	}
	
//	public Direction OppositeDirection(Direction dir){
//		switch(dir){
//		case LEFT:
//			return Direction.RIGHT;
//		case RIGHT:
//			return Direction.LEFT;
//		case UP:
//			return Direction.DOWN;
//		case DOWN:
//			return Direction.UP;
//		default:
//			return Direction.NULL;
//		}
//	}
	
}
